(* Reverse polish notation calculator. *)

%{
  val yyvalue = ref (0.0 : num);
%}

%token NUM

%% (* Grammar rules and actions follow *)
input:    (* empty *)
        | input line
;

line:     '\n'
| exp '\n'  { pretty stdout (pack $1); #flush stdout (); $$ 0.0 (* dummy value *) }
;

exp:      NUM
        | exp exp '+'     { $$ ($1 #+ $2)    }
        | exp exp '-'     { $$ ($1 #- $2)    }
        | exp exp '*'     { $$ ($1 #* $2)    }
        | exp exp '/'     { $$ ($1 #/ $2)    }
      (* Exponentiation *)
        | exp exp '^'     { $$ ($1 #^ $2) }
      (* Unary minus    *)
        | exp 'n'         { $$ (#~ $1)        }
;
%%

(* Lexical analyzer returns a double floating point 
   number on the stack and the token NUM, or the ASCII
   character read if not a number.  Skips all blanks
   and tabs, returns 0 for EOF. *)

val yyline = ref "";
val yypos = ref 0;

fun num_value (s, start, stop) = (* return the numerical value of a string obeying the regular
				 expression defined in yymatch_num below. *)
    let val pos = ref start
        fun int_value acc =
            if !pos<stop
               then let val c = ordof (s, !pos)
                    in
                       if c>=48 (* '0' *) andalso c<=57 (* '9' *)
                          then (inc pos;
                                int_value (10.0 #* acc #+ num (c-48)))
                       else acc
                    end
            else acc
        fun signed_int_value () =
            if !pos<stop andalso ordof (s, !pos)=45 (* '-' *)
               then (inc pos; #~ (int_value 0.0))
            else int_value 0.0
        fun frac_value () =
            if !pos<stop
               then let val c = ordof (s, !pos)
                    in
                       if c>=48 (* '0' *) andalso c<=57 (* '9' *)
                          then (inc pos;
                                0.1 #* (frac_value () #+ num (c-48)))
                       else 0.0
                    end
            else 0.0
        fun real_value () =
            let val mantissa = int_value 0.0
            in
               if !pos<stop andalso ordof (s, !pos)=46 (* '.' *)
                  then (inc pos; mantissa #+ frac_value ())
               else mantissa
            end
        fun real_exp_value () =
	    let val base = real_value ()
            in
               if !pos<stop andalso ordof (s, !pos)=69 (* 'E' *)
                  then (inc pos; base #* 10.0 #^ signed_int_value ())
               else base
            end
    in
       if !pos<stop andalso ordof (s, !pos)=45 (* '-' *)
          then (inc pos; #~ (real_exp_value ()))
       else real_exp_value ()
    end;

val yymatch_blank = #matchsub (regexp "^[ \t]*");
val yymatch_num = #matchsub (regexp "^-?[0-9]+(\\.[0-9]+)?(E-?[0-9]+)?");

fun yylex () =
    let val line = !yyline
        val stop = size line
        val SOME (_, pos) = yymatch_blank (line, !yypos, stop)
    in
       if pos=stop (* end of line *)
          then case #getline stdin () of (* load next line *)
                   "" => 0 (* end-of-file *)
                 | s => (yyline := s; yypos := 0; yylex ())
       else case yymatch_num (line, pos, stop) of
                SOME (_, pos') => (yyvalue := num_value (line, pos, pos'); yypos := pos'; NUM)
                                  (* this is a number, return NUM and put its value in yyvalue. *)
              | NONE => (yypos := pos+1; ordof (line, pos))
                        (* otherwise, return first non-blank character. *)
    end;

fun yyerror tokens = raise Parse tokens;

val yyparse_data = hyacc_data ((), yylex, 0.0, yyvalue, iarray (4, 0), yyerror);

fun main () = yyparse yyparse_data;
